(******************************************************************
 *                                                                *
 * Cryptographic protocol verifier                                *
 *                                                                *
 * Bruno Blanchet, David Cadé, Charlie Jacomme, and Benjamin Lipp *
 *                                                                *
 * Copyright (C) ENS, CNRS, INRIA, MPI-SP, 2005-2023              *
 *                                                                *
 ******************************************************************)

(*

    Copyright ENS, CNRS, INRIA, MPI-SP
    contributors: Bruno Blanchet, Bruno.Blanchet@inria.fr
                  David Cadé
                  Charlie Jacomme, Charlie.Jacomme@inria.fr
                  Benjamin Lipp, benjamin.lipp@mpi-sp.org

This software is a computer program whose purpose is to verify 
cryptographic protocols in the computational model.

This software is governed by the CeCILL-B license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-B
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-B license and that you accept its terms.

*)
open Types

val check_type : Parsing_helper.extent -> term -> typet -> unit
val check_bit_string_type : Parsing_helper.extent -> typet -> unit
val check_type_list :
  Parsing_helper.extent -> ('a * Parsing_helper.extent) list ->
  term list -> typet list -> unit
val check_array_type_list :
  Parsing_helper.extent -> ('a * Parsing_helper.extent) list ->
  term list -> repl_index list -> repl_index list -> term list

val dummy_if_fun : funsymb
val get_if_fun_tl : Parsing_helper.extent -> term list -> funsymb
    
module StringMap : Map.S with type key = string

(* Environment.
   May contain function symbols, variables, ...
   Is a map from strings to the description of the ident *)

type env_entry =
    EFunc of funsymb
  | EEvent of funsymb
  | EParam of param
  | EProba of proba
  | EType of typet
  | EVar of binder
  | EReplIndex of repl_index
  | EChannel of channel
  | ELetFun of funsymb * env_type * (Ptree.ident * Ptree.ty(*type*)) list * Ptree.term_e
  | EProcess of env_type * (Ptree.ident * Ptree.ty(*type*)) list * Ptree.process_e
  | ETable of table
  | EVarProba of var_proba
  | ELetProba of proba * env_type * (Ptree.ident * Ptree.dim_e(*dimension*)) list * (env_type -> probaf)

and env_type = env_entry StringMap.t

val env : env_type ref

val decl_name : env_entry -> string
    
val get_param : env_type -> string -> Parsing_helper.extent -> param
val get_event : env_type -> string -> Parsing_helper.extent -> funsymb
val get_type : env_type -> string -> Parsing_helper.extent -> typet
val get_type_or_param : env_type -> string -> Parsing_helper.extent -> typet
val get_ty : env_type -> Ptree.ty(*type*) -> typet * Parsing_helper.extent
val get_process : env_type -> string -> Parsing_helper.extent ->
  env_type * (Ptree.ident * Ptree.ty(*type*)) list * Ptree.process_e
val get_table : env_type -> string -> Parsing_helper.extent -> table
val get_function_no_letfun_if_allowed : env_type -> string * Parsing_helper.extent ->
  ('a * Parsing_helper.extent) list -> term list -> Parsing_helper.extent -> funsymb

(* The special polymorphic "if_fun" function is not treated by the next two functions.
   It must be rejected by the caller *)
val get_function_no_letfun_no_if : env_type -> string -> Parsing_helper.extent -> funsymb
val get_function_or_letfun_no_if : env_type -> string -> Parsing_helper.extent -> funsymb

(* Functions for dimensions *)

val dim_list_to_string : dim list -> string
val proba_dim_list_to_string : ('a * computed_dim) list -> string
    
val time_dim : computed_dim
val length_dim : computed_dim
val proba_dim : computed_dim
val num_dim : computed_dim

val apply_proba : ident -> env_type -> (probaf * computed_dim) list -> probaf
    
(* Global binder environment *)

type err_mess

val error_ambiguous : err_mess
val error_no_type : err_mess
val error_find_cond : err_mess
val error_in_input_process : err_mess

type binder_env_type

val empty_binder_env : binder_env_type
      
val add_in_env1 : binder_env_type -> string -> typet -> repl_index list -> binder_env_type
val add_in_env1reusename : binder_env_type -> string -> binder -> typet -> repl_index list -> binder_env_type
val add_in_env1error : binder_env_type -> err_mess -> string -> binder_env_type
val add_in_env1existing : binder_env_type -> string -> binder -> binder_env_type
val union_both : binder_env_type -> binder_env_type -> binder_env_type
val union_exclude : binder_env_type -> binder_env_type -> binder_env_type

val set_binder_env : binder_env_type -> unit
val set_and_get_old_binder_env : binder_env_type -> binder_env_type
exception Undefined of ident
val get_global_binder : string -> ident -> binder
val get_global_binder_if_possible : string -> binder option

val cst_for_type : typet -> term
val make_cst : term -> bool * term
